PallyPower = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceDB-2.0","AceEvent-2.0","AceDebug-2.0", "AceComm-2.0");

local dewdrop = AceLibrary("Dewdrop-2.0");
local RL = AceLibrary("Roster-2.1");
local L = AceLibrary("AceLocale-2.2"):new("PallyPower");

local classlist, classes = {}, {};
LastCast = {};
PallyPower_Assignments = {};
AllPallys = {};

local initalized = false
PP_Symbols = 0;
PP_IsPally = false;

function PallyPower:OnInitialize()
    self:RegisterDB("PallyPowerDB");
    self:RegisterChatCommand({"/pp"}, self.options);
    self:RegisterDefaults("profile", PALLYPOWER_DEFAULT_VALUES)
    self.player = UnitName("player");
    self.opt = self.db.profile;
    self:ScanSpells();
    self:ScanInventory();
    self:CreateLayout();
    dewdrop:Register(PallyPowerConfigFrame, "children", 
        function(level, value) dewdrop:FeedAceOptionsTable(self.options) end,
        "dontHook", true
    )
end

function PallyPower:OnEnable()
    -- events
    self:RegisterEvent("CHAT_MSG_ADDON");
    self:RegisterEvent("CHAT_MSG_SYSTEM")
    self:RegisterEvent("PLAYER_REGEN_ENABLED");
    self:RegisterEvent("SPELLS_CHANGED");
    self:RegisterEvent("RosterLib_RosterUpdated", "UpdateRoster")
    self:ScheduleRepeatingEvent("PallyPowerInventoryScan", self.InventoryScan, 60, self);
    self:UpdateRoster();
end

function PallyPower:OnDisable()
    -- events
    for i = 1, PALLYPOWER_MAXCLASSES do
        classlist[i] = 0;
        classes[i] = {};
    end
    self:UpdateLayout();
end

--
--  Config Window functionality
--

function PallyPowerConfig_Clear()
    PallyPower:ClearAssignments(UnitName("player"));
    if PallyPower:CheckRaidLeader(UnitName("player")) then
	PallyPower:SendMessage("CLEAR");
    end
end

function PallyPowerConfig_Options()

end

function PallyPower:Reset()
    self.opt.buffscale = 0.90;
    self.opt.configscale = 0.90;
    self.opt.smartbuffs = true;
    self.opt.display.rows = 1;
    self.opt.display.columns = 9;
    self.opt.display.gapping = 0;
    self.opt.display.buttonWidth = 100;
    self.opt.display.buttonHeight = 34;
    self.opt.display.alignClassButtons = "9";
    self.opt.display.alignPlayerButtons = "left";
    self:UpdateLayout();
end

function PallyPowerConfig_Refresh()
    AllPallys = {};
    PallyPower:ScanSpells();
    PallyPower:ScanInventory();
    PallyPower:SendSelf();
    PallyPower:SendMessage("REQ");
    PallyPower:UpdateLayout();
end

function PallyPowerConfig_Toggle(msg)
    if PallyPowerConfigFrame:IsVisible() then
	PallyPowerConfigFrame:Hide()
    else
	PallyPowerConfigFrame:Show()
    end
end

function PallyPowerConfig_ShowCredits()
   GameTooltip:SetOwner(this, "ANCHOR_TOPLEFT")
   GameTooltip:SetText(PallyPower_Credits1, 1, 1, 1)
--   GameTooltip:AddLine(PallyPower_Credits2, 1, 1, 1);
--   GameTooltip:AddLine(PallyPower_Credits3);
--   GameTooltip:AddLine(PallyPower_Credits4, 0, 1 ,0);
--   GameTooltip:AddLine(PallyPower_Credits5);
   GameTooltip:Show()
end

function PallyPowerGridButton_OnClick(btn, mouseBtn)
    _, _, pnum, class = string.find(btn:GetName(), "PallyPowerConfigFramePlayer(.+)Class(.+)");
    pnum = pnum + 0;
    class = class + 0;
    pname = getglobal("PallyPowerConfigFramePlayer"..pnum.."Name"):GetText();
    if not PallyPower:CanControl(pname) then return false end
   
    if (mouseBtn == "RightButton") then
	PallyPower_Assignments[pname][class] = 0;
	PallyPower:SendMessage("ASSIGN "..pname.." "..class.. " 0");
    else
	PallyPower:PerformCycle(pname, class);
    end
end

function PallyPowerGridButton_OnMouseWheel(btn, arg1)
    _,_,pnum,class = string.find(btn:GetName(), "PallyPowerConfigFramePlayer(.+)Class(.+)");
    pnum = pnum + 0;
    class = class + 0;
    pname = getglobal("PallyPowerConfigFramePlayer"..pnum.."Name"):GetText()
    if not PallyPower:CanControl(pname) then return false end

    if (arg1==-1) then  --mouse wheel down
        PallyPower:PerformCycle(pname, class)    
    else
        PallyPower:PerformCycleBackwards(pname, class)
    end     
end

function PallyPowerConfigFrame_MouseUp()
    if ( PallyPowerConfigFrame.isMoving ) then 
    	PallyPowerConfigFrame:StopMovingOrSizing(); 
	PallyPowerConfigFrame.isMoving = false; 
    end
end

function PallyPowerConfigFrame_MouseDown(arg1)
    if ( ( ( not PallyPowerConfigFrame.isLocked ) or ( PallyPowerConfigFrame.isLocked == 0 ) ) and ( arg1 == "LeftButton" ) ) then 
	PallyPowerConfigFrame:StartMoving(); 
	PallyPowerConfigFrame.isMoving = true; 
    end
end

function PallyPowerConfigGrid_Update()
    if not initalized then PallyPower:ScanSpells() end
    local i = 1;
    local numPallys = 0;
    local name, skills;
    if PallyPowerConfigFrame:IsVisible() then
	PallyPowerConfigFrame:SetScale(PallyPower.opt.configscale);
	for name in pairs(AllPallys) do
	    local fname = "PallyPowerConfigFramePlayer" .. i;

	    local SkillInfo = AllPallys[name];
	    local BuffInfo = PallyPower_Assignments[name];
	    getglobal(fname .. "Name"):SetText(name);

	    if PallyPower:CanControl(name) then
	        getglobal(fname.."Name"):SetTextColor(1,1,1);
            else
		if PallyPower:CheckRaidLeader(name) then
        	    getglobal(fname.."Name"):SetTextColor(0,1,0);
		else
        	    getglobal(fname.."Name"):SetTextColor(1,0,0);
		end
	    end
	    getglobal(fname .. "Symbols"):SetText(SkillInfo.symbols);
	    getglobal(fname .. "Symbols"):SetTextColor(1,1,0.5);
	    for id = 1, 6 do
		if SkillInfo[id] then
		    getglobal(fname.."Icon"..id):Show();
		    getglobal(fname.."Skill"..id):Show();
		    local txt = SkillInfo[id].rank;
		    if SkillInfo[id].talent + 0 > 0 then 
			txt = txt.. "+" .. SkillInfo[id].talent;
		    end
		    getglobal(fname.."Skill"..id):SetText(txt);
		else
		    getglobal(fname.."Icon"..id):Hide();
		    getglobal(fname.."Skill"..id):Hide();
		end
	    end
	    for id = 1, PALLYPOWER_MAXCLASSES do
		if BuffInfo and BuffInfo[id] then
		    getglobal(fname.."Class"..id.."Icon"):SetTexture(PallyPower.BlessingIcons[BuffInfo[id]]);
		else
		    getglobal(fname.."Class"..id.."Icon"):SetTexture(nil);
		end                
	    end
	    i = i + 1;
	    numPallys = numPallys + 1;
	end
	PallyPowerConfigFrame:SetHeight(14 + 24 + 56 + (numPallys * 56) + 22);
	for i = 1, 12 do
	    local fname = "PallyPowerConfigFramePlayer" .. i;
	    if i <= numPallys then
		getglobal(fname):Show();
	    else
		getglobal(fname):Hide();
	    end
	end
    end
end

--
-- Main functionality
--

function PallyPower:Report(type)
    if self:GetNumUnits() > 0 then
	if not type then
	    if GetNumRaidMembers() > 0 then
      	    	type = "RAID";
	    else	
	    	type = "PARTY";
	    end
	end
    	if PallyPower:CheckRaidLeader(self.player) then
    	    SendChatMessage(PALLYPOWER_ASSIGNMENTS1, type)
	    local list = {};
    	    for name in pairs(AllPallys) do
	    	local blessings;
	    	for i = 1, 6 do
		    list[i] = 0;
	    	end
	    	for id = 1, PALLYPOWER_MAXCLASSES do
	            local bid = PallyPower_Assignments[name][id];
	            if bid and bid > 0 then
	    	    	list[bid] = list[bid] + 1;
            	    end
	    	end
	        for id = 1, 6 do
	            if (list[id] > 0) then
 	  	    	if (blessings) then 
            	            blessings = blessings .. ", "; 
          	    	else
            	            blessings = "";
          	    	end
	            	local _,_, spell = string.find(PallyPower.Spells[id], PallyPower_BlessingNameSearch);
                    	blessings = blessings .. spell;
                    end
      	    	end
      	    	if not (blessings) then
	            blessings = "Nothing";
      	    	end
      	    	SendChatMessage(name ..": ".. blessings, type)
	    end
            SendChatMessage(PALLYPOWER_ASSIGNMENTS2, type)
	else
	    self:Print(ERR_NOT_LEADER);
    	end
    else
	self:Print(ERR_NOT_IN_RAID);
    end
end

function PallyPower:PerformCycle(name, class)
    shift = IsShiftKeyDown();

    if shift then class = 4 end
	
    if not PallyPower_Assignments[name] then
	PallyPower_Assignments[name] = { };
    end

    if not PallyPower_Assignments[name][class] then 
	cur=0;
    else 
    	cur=PallyPower_Assignments[name][class];
    end

    PallyPower_Assignments[name][class] = 0;
    
    for test = cur+1, 7 do
	if PallyPower:CanBuff(name, test) and (PallyPower:NeedsBuff(class, test) or shift) then
		cur = test
      		do break end
    	end
    end
   
    if cur == 7 then cur = 0 end

    if shift then
    	for test = 1, PALLYPOWER_MAXCLASSES do
            PallyPower_Assignments[name][test] = cur
        end
    	PallyPower:SendMessage("MASSIGN "..name.." "..cur)
    else
    	PallyPower_Assignments[name][class] = cur
    	PallyPower:SendMessage("ASSIGN "..name.." "..class.." "..cur)  
    end
end

function PallyPower:PerformCycleBackwards(name, class)
    shift=IsShiftKeyDown();

    if shift then class=4 end  

    if not PallyPower_Assignments[name] then
	PallyPower_Assignments[name] = { };
    end

    if not PallyPower_Assignments[name][class] then 
	cur=7;
    else 
	cur=PallyPower_Assignments[name][class];
	if cur == 0 then cur = 7 end
    end

    PallyPower_Assignments[name][class] = 0;

    for test = cur-1, 0, -1 do
	cur = test
	if PallyPower:CanBuff(name, test) and (PallyPower:NeedsBuff(class, test) or shift) then
      		do break end
	end
    end

    if shift then
    	for test = 1, PALLYPOWER_MAXCLASSES do
	    PallyPower_Assignments[name][test] = cur;
	end
	PallyPower:SendMessage("MASSIGN "..name.." "..cur)
    else
	PallyPower_Assignments[name][class] = cur;
	PallyPower:SendMessage("ASSIGN "..name.." "..class.." "..cur);
    end
end

function PallyPower:CanBuff(name, test)
    if test==7 then 
        return true 
    end

    if (not AllPallys[name][test]) or (AllPallys[name][test].rank == 0) then 
        return false 
    end
    return true
end

function PallyPower:NeedsBuff(class, test)
    if test==7 or test==0 then 
    	return true 
    end

    if self.opt.smartbuffs then
	-- no wisdom for warriors and rogues
    	if (class == 1 or class==2) and test == 1 then 
	    return false
	end
   	-- no salv for warriors
        if class == 1 and test == 3 then 
          return false 
        end
    	-- no might for casters
        if (class == 3 or class == 7 or class == 8) and test == 2 then
          return false
        end
    end

    for name, skills in pairs(PallyPower_Assignments) do
	if (AllPallys[name]) and ((skills[class]) and (skills[class]==test)) then 
    	    return false 
	end
    end
    return true
end

function PallyPower:ScanSpells()
    self:Debug("Scan Spells -- begin");
    _, class=UnitClass("player");
    if (class == "PALADIN") then
	local RankInfo = {};
        local i = 1;
        while true do
            local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
            --local spellTexture = GetSpellTexture(i, BOOKTYPE_SPELL)
            if not spellName then
                do break end
            end
    
            if not spellRank or spellRank == "" then 
		spellRank = PallyPower_Rank1
            end
            local _,_,bless = string.find(spellName, PallyPower_BlessingSearch)
	    if bless then
                for id, name in pairs(PallyPower.Spells) do
                    if name == bless then
                        local _,_,rank = string.find(spellRank, PallyPower_RankSearch);
                        if (RankInfo[id] and spellRank < RankInfo[id].rank) then
                            -- Do Nothing
                        else
                            RankInfo[id] = {};
                            RankInfo[id].rank = rank;
                            RankInfo[id].talent = 0;
                        end
			break;
                    end
                end
            end
            i = i + 1
        end	
        local numTabs = GetNumTalentTabs();
        for t= 1, numTabs do
            local numTalents = GetNumTalents(t);
                for i = 1, numTalents do
                  nameTalent, icon, iconx, icony, currRank, maxRank = GetTalentInfo(t, i);
                  local _,_,bless = string.find(nameTalent, PallyPower_TalentSearch) 
                  if bless then
                      initalized=true;
                      for id, name in pairs(PallyPower.Spells) do
                          if name == bless then
                              if (RankInfo[id]) then RankInfo[id].talent = currRank;
			      break;
                          end
                      end
                   end
                end              
            end
        end
	AllPallys[self.player] = RankInfo;
        PP_IsPally = true;
    else
      	PP_IsPally = false;
        initalized=true;
    end
    self:Debug("Scan Spells -- end");
end

function PallyPower:ScanInventory()
    self:Debug("Scan Inventory -- begin");
    if not PP_IsPally then return end
    local num = 0;
    for bag = 0,  4 do
        local bagslots = GetContainerNumSlots(bag);
        if (bagslots) then
            for slot = 1, bagslots do
                local link = GetContainerItemLink(bag, slot)
	        if (link and string.find(link, PallyPower_Symbol)) then
	            local _, count, locked = GetContainerItemInfo(bag, slot);
	            num = num + count;
                end
            end
        end
    end
    if PP_Symbols ~= num then
	PP_Symbols = num;
    end
    AllPallys[self.player].symbols = PP_Symbols;
    self:Debug("Scan Inventory -- end");
end

function PallyPower:InventoryScan()
    self:ScanInventory();
    if self:GetNumUnits() > 0 then
	self:SendMessage("SYMCOUNT " .. PP_Symbols);
    end
end

function PallyPower:SendSelf()
    self:Debug("Send self -- begin");
    if not initalized then PallyPower:ScanSpells() end
    if not AllPallys[self.player] then return end
--    local name = UnitName("player");
    local s;

    local SkillInfo = AllPallys[self.player];
    s = "";
    for i = 1, 6 do
        if not SkillInfo[i] then
	    s = s.."nn";
	else
	    s = s .. SkillInfo[i].rank .. SkillInfo[i].talent;
	end
    end
    s = s .. "@";

    if not PallyPower_Assignments[self.player] then
        PallyPower_Assignments[self.player] = {};
	for i = 1, PALLYPOWER_MAXCLASSES do
	    PallyPower_Assignments[self.player][i] = 0;
	end
    end	

    local BuffInfo = PallyPower_Assignments[self.player];

    for i = 1, PALLYPOWER_MAXCLASSES do
	if not BuffInfo[i] or BuffInfo[i] == 0 then
	    s = s .. "n";
	else
	    s = s .. BuffInfo[i];
	end
    end

    self:SendMessage("SELF " .. s);
    self:SendMessage("SYMCOUNT " .. PP_Symbols);
    self:Debug("Send self -- end");
end

function PallyPower:SendMessage(msg)
    self:Debug("Sending message");
    local type;
    if GetNumRaidMembers() == 0 then
	type = "PARTY";
    else
	type = "RAID";
    end
    SendAddonMessage(PallyPower.commPrefix, msg, type, self.player);
end

function PallyPower:SPELLS_CHANGED()
    self:ScanSpells();
end

function PallyPower:CHAT_MSG_ADDON(prefix, message, distribution, sender)
    self:Debug("CHAT_MSG_ADDON event");
    if prefix == PallyPower.commPrefix and (distribution == "PARTY" or distribution == "RAID") then
	self:ParseMessage(sender, message);
    end
end

function PallyPower:CHAT_MSG_SYSTEM()
    self:Debug("CHAT_MSG_SYSTEM event");
    if string.find(arg1, ERR_RAID_YOU_JOINED) then
	self:SendSelf();
	self:SendMessage("REQ");
    end
end

function PallyPower:PLAYER_REGEN_ENABLED()
    if PP_IsPally then self:UpdateLayout() end
end

function PallyPower:CanControl(name)
    return (IsPartyLeader() or IsRaidLeader() or IsRaidOfficer() or (name==self.player));
end

function PallyPower:CheckRaidLeader(nick)
    if GetNumRaidMembers() == 0 then
	for i = 1, GetNumPartyMembers() do
	    if nick == UnitName("party"..i) and UnitIsPartyLeader("party"..i) then
		return true;
	    end
	end
	if nick == self.player and UnitIsPartyLeader("player") then
	    return true;
	end
	return false;
    else
    	for i = 1, GetNumRaidMembers() do
	    local name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(i);
	    if nick == name and rank >= 1 then
		return true;
	    end
	end
	return false;
    end
end

function PallyPower:ClearAssignments(sender)
    local leader = self:CheckRaidLeader(sender);
    for name, skills in pairs(PallyPower_Assignments) do
	if leader or name == sender then
	    --self:Print("Clearing: %s", name);
	    for i = 1, PALLYPOWER_MAXCLASSES do
		PallyPower_Assignments[name][i] = 0;
	    end
	end
    end
end

function PallyPower:ParseMessage(sender, msg)
--    self:Print("Received from: %s, message: %s", sender, msg);
    if sender ~= self.player then
	local leader = self:CheckRaidLeader(sender);

	if msg == "REQ" then
	    self:SendSelf();
	end

	if string.find(msg, "^SELF") then
	    PallyPower_Assignments[sender] = { };
	    AllPallys[sender] = { };
	    _, _, numbers, assign = string.find(msg, "SELF ([0-9n]*)@([0-9n]*)")
	    for i = 1, 6 do
		rank = string.sub(numbers, (i - 1) * 2 + 1, (i - 1) * 2 + 1);
                talent = string.sub(numbers, (i - 1) * 2 + 2, (i - 1) * 2 + 2);
		if rank ~= "n" then
		    AllPallys[sender][i] = { };
		    AllPallys[sender][i].rank = rank;
		    AllPallys[sender][i].talent = talent;
		end
	    end
	    if assign then
		for i = 1, PALLYPOWER_MAXCLASSES do
		    tmp =string.sub(assign, i, i);
		    if tmp == "n" or tmp == "" then tmp = 0 end
		    PallyPower_Assignments[sender][i] = tmp + 0;
		end
	    end
	end

	if string.find(msg, "^ASSIGN") then
	    _, _, name, class, skill = string.find(msg, "^ASSIGN (.*) (.*) (.*)")
	    if not name == sender and not leader then return false end;
	    if not PallyPower_Assignments[name] then PallyPower_Assignments[name] = {} end;
	    class = class + 0;
	    skill = skill + 0;
	    PallyPower_Assignments[name][class] = skill;	    
	end

	if string.find(msg, "^MASSIGN") then
	    _, _, name, skill = string.find(msg, "^MASSIGN (.*) (.*)")
	    if not name == sender and not leader then return false end;
	    if not PallyPower_Assignments[name] then PallyPower_Assignments[name] = {} end;
	    skill = skill + 0;
	    for i = 1, PALLYPOWER_MAXCLASSES do
	        PallyPower_Assignments[name][i] = skill;
	    end
	end

	if string.find(msg, "^SYMCOUNT") then
            _, _, count = string.find(msg, "^SYMCOUNT ([0-9]*)")
            if AllPallys[sender] then
                AllPallys[sender].symbols = count;
	    else
                self:SendMessage("REQ");
	    end
	end

	if string.find(msg, "^CLEAR") then
	    if leader then
	        self:ClearAssignments(sender);
	    end
	end

    end
end

function PallyPower:FormatTime(time)
    if not time or time < 0 then 
    	return ""; 
    end
    mins = floor(time / 60)
    secs = time - (mins * 60)
    return string.format("%d:%02d", mins, secs);
end

function PallyPower:OnUpdate()
    for i,k in pairs(LastCast) do LastCast[i] = k - arg1 end
end

function PallyPower:GetClassID(class)
    for id, name in pairs(self.ClassID) do
        if (name==class) then
            return id
        end
    end
    return -1
end

function PallyPower:ShouldIDisplay()
    if GetNumRaidMembers() > 0 then 
        return true 
    end
    if GetNumPartyMembers() > 0 and self.opt.ShowInParty then 
        return true 
    end
    return false
end

function PallyPower:GetNumUnits()
    if GetNumRaidMembers() > 0 then 
        return GetNumRaidMembers() 
    end
    if GetNumPartyMembers() > 0 and self.opt.ShowInParty or self.opt.ShowWhenSingle then
        return GetNumPartyMembers() + 1
    end
    return 0
end

function PallyPower:RosterLib_UnitChanged(unitid, name, class, subgroup, rank, oldname, oldunitid, oldclass, oldsubgroup, oldrank)
    if not name then
	self:Debug("UnitLeft", unitid, name, class, subgroup, rank, oldname, oldunitid, oldclass, oldsubgroup, oldrank)
    elseif not oldname then
	self:Debug("UnitJoined", unitid, name, class, subgroup, rank, oldname, oldunitid, oldclass, oldsubgroup, oldrank)
    else
	--nothing
    end
end

function PallyPower:UpdateRoster()
    -- unregister events
    self:Debug("Update Roster");
    self:CancelScheduledEvent("PallyPowerUpdateButtons");

    local num = self:GetNumUnits();
          
    for i = 1, PALLYPOWER_MAXCLASSES do
        classlist[i] = 0;
        classes[i] = {};
    end

    if num > 0 and PP_IsPally then
    	for unit in RL:IterateRoster(true) do
            for i = 1, PALLYPOWER_MAXCLASSES do
            	if unit.class == self.ClassID[i] then
                    local tmp = unit;
		    tmp.visible = false;
		    tmp.hasbuff = false;
		    tmp.dead = false;
                    classlist[i] = classlist[i] + 1;
		    table.insert(classes[i], tmp);
             	end
            end
    	end
    end

    self:UpdateLayout();

    if num > 0 and PP_IsPally then 
	-- register events
	self:ScheduleRepeatingEvent("PallyPowerUpdateButtons", self.ButtonsUpdate, 2.0, self);
    end	
    self:Debug("Update Roster - end");
end

function PallyPower:ScanClass(classID)
    --    self:Print("Scanning class: %s -- begin", classID);

    local class = classes[classID];
    local spell = PallyPower:GetSpell(classID);

    for playerID, unit in pairs(class) do
	if unit.unitid then
	    if IsSpellInRange(spell, unit.unitid) == 1 then
		unit.visible = true;
	    else
		unit.visible = false;
	    end
	    if UnitIsDeadOrGhost(unit.unitid) then
		unit.dead = true;
	    else
		unit.dead = false;
	    end
	    if self:IsBuffActive(spell, unit.unitid) then
		unit.hasbuff = true;
	    else
		unit.hasbuff = false;
	    end
	end
    end
end

function PallyPower:CreateLayout()
    self:Debug("Create Layout -- begin");

    local p = _G["PallyPowerHeader"];
    -- State 0 shows the class buttons
    p:SetAttribute("statebindings", "0:top;1-"..PALLYPOWER_MAXCLASSES..":sub");
    p:SetAttribute("statemap-anchor", "0-"..PALLYPOWER_MAXCLASSES..":$input");
    p:SetAttribute("delaystatemap-anchor", "0-"..PALLYPOWER_MAXCLASSES..":0");
    p:SetAttribute("delaytimemap-anchor", "0-"..PALLYPOWER_MAXCLASSES..":0.5");
    p:SetAttribute("delayhovermap-anchor", "0-"..PALLYPOWER_MAXCLASSES..":true");

    self.Header = p;

    self.classButtons = {};
    self.playerButtons = {};

    for cbNum = 1, PALLYPOWER_MAXCLASSES do 
	-- create class buttons
        local cButton = CreateFrame("Button", "PallyPowerSAB" .. cbNum, self.Header, "PallyPowerSAButtonTemplate, SecureAnchorEnterTemplate");
        self.classButtons[cbNum] = cButton;
	cButton:RegisterForClicks("LeftButtonDown", "RightButtonDown");
	cButton:SetAttribute("newstate", tostring(cbNum));
        cButton:SetAttribute("anchorchild", self.Header);
        cButton:SetAttribute("childstate", "^" .. tostring(cbNum));
        cButton:SetAttribute("childraise", true);
        self.Header:SetAttribute("addchild", cButton);

	-- create player buttons
	self.playerButtons[cbNum] = {};
        local pButtons = self.playerButtons[cbNum];
	
        for pbNum = 1, PALLYPOWER_MAXPERCLASS do -- create player buttons for each class
            local pButton = CreateFrame("Button","PallyPowerSAB".. cbNum .. "P" .. pbNum, self.Header, "PallyPowerSAPopupButtonTemplate");
            pButtons[pbNum] = pButton;
	    --pButton:SetAttribute("newstate", 0);
   	    pButton:RegisterForClicks("LeftButtonDown", "RightButtonDown");
	    self.Header:SetAttribute("anchorchild", pButton);
            self.Header:SetAttribute("addchild", pButton);
        end -- by pbNum
    end -- by classIndex

    self:UpdateLayout();
    self:Debug("Create Layout -- end");
end

function PallyPower:CountClasses()
    local val = 0;
    if not classes then return 0 end;
    for i = 1, PALLYPOWER_MAXCLASSES do
	if classlist[i] and classlist[i] > 0 then
	    val = val + 1;
	end
    end
    return val;
end

function PallyPower:UpdateLayout()
    self:Debug("Update Layout -- begin");
    if InCombatLockdown() then return false end;
    PallyPowerFrame:SetScale(self.opt.buffscale);
    local rows = self.opt.display.rows;
    local columns = self.opt.display.columns;
    local gapping = self.opt.display.gapping;
    local buttonWidth = self.opt.display.buttonWidth;
    local buttonHeight = self.opt.display.buttonHeight;
    local centerShiftX = 0;
    local centerShiftY = 0;
    local point = "BOTTOMLEFT";
    local x = (buttonWidth + gapping);
    local y = (buttonHeight + gapping);
    local displayedButtons = math.min(self:CountClasses(),rows, columns);
    local displayedColumns = math.min(displayedButtons, columns);
    local displayedRows = math.floor((displayedButtons - 1) / columns) + 1;

    if (self.opt.display.alignClassButtons == "1") then
	point = "BOTTOMLEFT";
    elseif (self.opt.display.alignClassButtons == "3") then
	x = x * -1;
	point = "BOTTOMRIGHT";
    elseif (self.opt.display.alignClassButtons == "7") then
	x = x * -1;
	y = y * -1;
	point = "TOPRIGHT";
    elseif (self.opt.display.alignClassButtons == "9") then
	y = y * -1;
	point = "TOPLEFT";
    end

    for cbNum = 1, PALLYPOWER_MAXCLASSES do -- position class buttons
	local cButton = self.classButtons[cbNum];
	-- set visual attributes
        self:SetButton("PallyPowerSAB" .. cbNum);
	-- set position
	cButton.x = (math.fmod(cbNum - 1, columns) * x + centerShiftX);
	cButton.y = math.floor((cbNum - 1) / columns) * y + centerShiftY;
        cButton:SetAttribute("ofsx", cButton.x);
        cButton:SetAttribute("ofsy", cButton.y);
	cButton:SetAttribute("ofspoint", "*:"..point);
	cButton:SetAttribute("ofsrelpoint", "*:".."CENTER");

        local hitRectOutside = -1 * math.ceil(math.max(buttonWidth, buttonHeight, 34, gapping) / 2);
        local hitRectInside = -1 * math.ceil(gapping / 2);
	cButton:SetHitRectInsets(hitRectOutside, hitRectOutside, hitRectInside, hitRectInside);

        local pButtons = self.playerButtons[cbNum];
        for pbNum = 1, PALLYPOWER_MAXPERCLASS do -- position player buttons
	    local pButton = pButtons[pbNum];
            self:SetPButton("PallyPowerSAB".. cbNum .. "P" .. pbNum) 
    	    --pButton:SetAttribute("showstates", tostring(cbNum));
	    pButton:SetAttribute("showstates", "!*");
    	    pButton:SetAttribute("ofspoint", "*:"..point);
    	    pButton:SetAttribute("ofsrelpoint", "*:".."CENTER");
            if (self.opt.display.alignPlayerButtons == "bottom") then
                pButton:SetAttribute("ofsx", "*:"..cButton.x);
                pButton:SetAttribute("ofsy", "*:"..(cButton.y - pbNum * (buttonHeight + gapping)));
            elseif (self.opt.display.alignPlayerButtons == "left") then
                pButton:SetAttribute("ofsx", "*:"..cButton.x - pbNum * (buttonWidth + gapping));
                pButton:SetAttribute("ofsy", "*:"..(cButton.y));
            elseif (self.opt.display.alignPlayerButtons == "right") then
                pButton:SetAttribute("ofsx", "*:"..cButton.x + pbNum * (buttonWidth + gapping));
                pButton:SetAttribute("ofsy", "*:"..(cButton.y));
	    elseif (self.opt.display.alignPlayerButtons == "top") then
                pButton:SetAttribute("ofsx", "*:"..cButton.x);
                pButton:SetAttribute("ofsy", "*:"..(cButton.y + pbNum * (buttonHeight + gapping)));
	    elseif (self.opt.display.alignPlayerButtons == "compact-right") then
		pButton:SetAttribute("ofsx", "*:"..cButton.x + (buttonWidth + gapping));
 	   	pButton:SetAttribute("ofsy", "*:"..cButton.y + (pbNum - 1) * (buttonHeight + gapping));
	    elseif (self.opt.display.alignPlayerButtons == "compact-left") then
		pButton:SetAttribute("ofsx", "*:"..cButton.x - (buttonWidth + gapping));
 	   	pButton:SetAttribute("ofsy", "*:"..cButton.y + (pbNum - 1) * (buttonHeight + gapping));
            end	
	end
    end

    local cbNum = 0;
    for classIndex = 1, PALLYPOWER_MAXCLASSES do
	if (classlist[classIndex] and classlist[classIndex] ~= 0 and PallyPower:GetSpellID(classIndex) ~= 0) then
	    cbNum = cbNum + 1;
	    local cButton = self.classButtons[cbNum];
            cButton:SetAttribute("classID", classIndex);
  	    cButton:SetAttribute("showstates", "*");
	    cButton:SetAttribute("type1", "spell");
	    cButton:SetAttribute("type2", "spell");
            self.Header:SetAttribute("addchild", cButton);
    	    local pButtons = self.playerButtons[cbNum];
	    for pbNum = 1, math.min(classlist[classIndex], PALLYPOWER_MAXPERCLASS) do
	    	local pButton = pButtons[pbNum];
	    	pButton:SetAttribute("classID", classIndex);
	    	pButton:SetAttribute("playerID", pbNum);
		if not self.opt.display.hidePlayerButtons then
            	    pButton:SetAttribute("showstates", tostring(cbNum));
		else
            	    pButton:SetAttribute("showstates", "!*");
		end
		local spell = self:GetSpell(classIndex)
		local gspell = L["SPELL_GTPREF"] .. spell .. L["SPELL_GTSUFF"];
		local unitid  = self:GetUnit(classIndex, pbNum);
	        -- left click (target a specific player and do 15 minute buff)
	    	pButton:SetAttribute("type1", "spell");
		pButton:SetAttribute("unit1", unitid);
		pButton:SetAttribute("spell1", gspell);
		-- right click (target a specific player and do 5 minute buff)
	    	pButton:SetAttribute("type2", "spell");
		pButton:SetAttribute("unit2", unitid);
		pButton:SetAttribute("spell2", spell);
	        self.Header:SetAttribute("addchild", pButton);
            end -- by pbnum
            for pbNum = classlist[classIndex]+1, PALLYPOWER_MAXPERCLASS do
            	local pButton = pButtons[pbNum];
	    	pButton:SetAttribute("classID", 0);
	    	pButton:SetAttribute("playerID", 0);
            	pButton:SetAttribute("showstates","!*");
	        self.Header:SetAttribute("addchild", pButton);

            end
    	end
    end
    cbNum = cbNum + 1;
    for i = cbNum, PALLYPOWER_MAXCLASSES do
	local cButton = self.classButtons[i];
	cButton:SetAttribute("classID", 0);
	cButton:SetAttribute("showstates", "!*");
        self.Header:SetAttribute("addchild", cButton);
 	local pButtons = self.playerButtons[cbNum];
        for pbNum = 1, PALLYPOWER_MAXPERCLASS do
	    local pButton = pButtons[pbNum];
	    pButton:SetAttribute("classID", 0);
	    pButton:SetAttribute("playerID", 0);
	    pButton:SetAttribute("showstates","!*");
            self.Header:SetAttribute("addchild", pButton);
	end
    end
    SecureStateHeader_Refresh(self.Header);

    self:ButtonsUpdate();
    self:UpdateAnchor(displayedButtons);
    self.Header:SetAttribute("state", "0");

    self:Debug("Update Layout -- end");
end

function PallyPower:SetButton(baseName)
--    self:Print("Set Button: %s -- begin", baseName);
    local time = _G[baseName.."Time"];
    local text = _G[baseName.."Text"];

    if (self.opt.display.HideCountText) then
	text:Hide();
    else
	text:Show();
    end

    if (self.opt.display.HideTimerText) then
	time:Hide();
    else
	time:Show();
    end
--    self:Print("Set Button -- end");
end

function PallyPower:SetPButton(baseName)
--    self:Print("Set PButton: %s -- begin", baseName);
    local rng = _G[baseName.."Rng"];
    local dead = _G[baseName.."Dead"];
    local name = _G[baseName.."Name"];

    if (self.opt.display.HideRngText) then
	rng:Hide();
    else
	rng:Show();
    end

    if (self.opt.display.HideDeadText) then
	dead:Hide();
    else
	dead:Show();
    end

    if (self.opt.display.HideNameText) then
	name:Hide();
    else
	name:Show();
    end
--    self:Print("Set PButton -- end");
end


function PallyPower:UpdateButton(button, baseName, classID)
--    self:Print("Update Button: %s, Class: %s", baseName, classID);
    local button = _G[baseName];
    local classIcon = _G[baseName.."ClassIcon"];
    local buffIcon = _G[baseName.."BuffIcon"];
    local time = _G[baseName.."Time"];
    local text = _G[baseName.."Text"];
    
    local nneed = 0;
    local nhave = 0;
    local ndead = 0;
    --self:Print("Scaninfo: %s", PP_ScanInfo[classID]);
    for playerID, unit in pairs(classes[classID]) do
	if unit.visible then
            if not unit.hasbuff then
		nneed = nneed + 1;
	    else
		nhave = nhave + 1;
	    end
	else
	    nhave = nhave + 1;
	end

	if unit.dead then
	    ndead = ndead + 1;
        end
    end
    classIcon:SetTexture(self.ClassIcons[classID]);
    classIcon:SetVertexColor(1, 1, 1);

    buffIcon:SetTexture(self.BlessingIcons[PallyPower:GetSpellID(classID)]);

    if InCombatLockdown() then
    	buffIcon:SetVertexColor(0.4, 0.4, 0.4);
    else
	buffIcon:SetVertexColor(1, 1, 1);
    end;

    time:SetText(self:FormatTime(LastCast[classID]));

    if ndead > 0 then
	text:SetText(nneed.." (" .. ndead ..")");
    else
	text:SetText(nneed);
    end

    if (nhave == 0) then
	button:SetBackdropColor(1.0, 0.0, 0.0, 0.5);
    elseif (nneed > 0) then
	button:SetBackdropColor(1.0, 1.0, 0.5, 0.5);
    else
        button:SetBackdropColor(0.0, 0.0, 0.0, 0.5);
    end	

--    self:Print("Update button -- end");
end

function PallyPower:UpdatePButton(button, baseName, classID, playerID)
--    self:Print("Update PButton: %s, Class: %s, Player: %s", baseName, classID, playerID);
    local button = _G[baseName];
    local buffIcon = _G[baseName.."BuffIcon"];
    local rng  = _G[baseName.."Rng"]; 
    local dead = _G[baseName.."Dead"];
    local name = _G[baseName.."Name"];

    local unit = classes[classID][playerID];
    if unit then
	local nneed = 0;
	local nhave = 0;
	local ndead = 0;

	if unit.visible then
	    if not unit.hasbuff then
		nneed = 1;
	    else
		nhave = 1;
	    end
	else
	    nhave = 1;
	end

	if unit.dead then
            ndead = 1;
	end

	buffIcon:SetTexture(self.BlessingIcons[PallyPower:GetSpellID(classID)]);
	buffIcon:SetVertexColor(1, 1, 1);

	if (nhave == 0) then
	    button:SetBackdropColor(1.0, 0.0, 0.0, 0.5);
	elseif (nneed == 1) then
	    button:SetBackdropColor(1.0, 1.0, 0.5, 0.5);
	else
            button:SetBackdropColor(0.0, 0.0, 0.0, 0.5);
	end	

	if unit.hasbuff then
	    buffIcon:SetAlpha(1);
	    if not unit.visible then
		rng:SetVertexColor(1, 0, 0);
		rng:SetAlpha(1);
	    else
		rng:SetVertexColor(0, 1, 0);
		rng:SetAlpha(1);
	    end
	    dead:SetAlpha(0);
        else
            buffIcon:SetAlpha(0);

	    if not unit.visible then
		rng:SetVertexColor(1, 0, 0);
		rng:SetAlpha(1);
	    else
		rng:SetVertexColor(0, 1, 0);
		rng:SetAlpha(1);
	    end

	    if unit.dead then
		dead:SetVertexColor(1, 0, 0);
		dead:SetAlpha(1);
	    else
		dead:SetVertexColor(0, 1, 0);
		dead:SetAlpha(0);
	    end;
	end;
	name:SetText(unit.name);
    else
        button:SetBackdropColor(0.0, 0.0, 0.0, 0.5);
	buffIcon:SetAlpha(0);
	rng:SetAlpha(0)
	dead:SetAlpha(0);
    end
--    self:Print("Update PopupButton -- end");
end

function PallyPower:ButtonsUpdate()
--    self:Print("ButtonsUpdate -- begin");
    local title = _G["PallyPowerAnchorText"];
    title:SetText(format(PallyPower.BuffBarTitle, PP_Symbols));
    for cbNum = 1, PALLYPOWER_MAXCLASSES do -- scan classes and if populated then assign textures, etc
        local cButton = self.classButtons[cbNum];
	local classIndex = cButton:GetAttribute("classID");
	if classIndex > 0 then
	    self:ScanClass(classIndex); -- scanning for in-range and buffs
	    self:UpdateButton(cButton, "PallyPowerSAB"..cbNum, classIndex)
	    local pButtons = self.playerButtons[cbNum];
            for pbNum = 1, PALLYPOWER_MAXPERCLASS do
                local pButton = pButtons[pbNum];
		local playerIndex = pButton:GetAttribute("playerID");
		if playerIndex > 0 then
		    self:UpdatePButton(pButton, "PallyPowerSAB".. cbNum .."P".. pbNum, classIndex, playerIndex)
		end
	    end -- by pbnum
        end -- class has players
    end  -- by cnum
--    self:Print("ButtonsUpdate -- end");
end

function PallyPower:UpdateAnchor(displayedButtons)
    PallyPowerFrame:SetFrameStrata("LOW");
    PallyPowerAnchor:SetFrameStrata("MEDIUM");
    PallyPowerAnchor:SetChecked(self.opt.display.frameLocked);
    PallyPowerAnchor:SetAttribute("state", "0");
    if (self.opt.display.hideDragHandle) then
	PallyPowerAnchor:SetAttribute("showstates", "!*");
	PallyPowerAnchor:Hide();
    else
	PallyPowerAnchor:SetAttribute("showstates", "*");
	PallyPowerAnchor:Show();
    end
end

function PallyPower:ButtonSetTooltip(button, elapsed)

end

function PallyPower:GetSpellID(classID)
    if PallyPower_Assignments[self.player] and PallyPower_Assignments[self.player][classID] then
	return PallyPower_Assignments[self.player][classID];
    else
	return 0;
    end 
end

function PallyPower:GetSpell(classID)
    return PallyPower.Spells[PallyPower:GetSpellID(classID)];
end

function PallyPower:GetUnit(classID, playerID)
    return classes[classID][playerID].unitid;
end

function PallyPower:GetUnitSmart(classID, mousebutton, spell)
    local i, class, unit;
    class = classes[classID];
    if (mousebutton == "LeftButton") then
	for i, unit in pairs(class) do
	    if IsSpellInRange(spell, unit.unitid) == 1 then
		return unit.unitid;
	    end
	end
    elseif (mousebutton == "RightButton") then
	for i, unit in pairs(class) do
	    if not self:IsBuffActive(spell, unit.unitid) and IsSpellInRange(spell, unit.unitid) == 1 then
		return unit.unitid;
	    end
	end
    end
    return nil
end

function PallyPower:IsBuffActive(spell, unitID)
    local result = false;
    local j = 1;
    while UnitBuff(unitID, j) do
        local name, _, _, _ = UnitBuff(unitID, j);
	if (name == spell) or (name == L["SPELL_GTPREF"] .. spell .. L["SPELL_GTSUFF"]) then
	    result = true;
	    break;
	end
	j = j + 1;
    end
    return result;
end

function PallyPower:ButtonPreClick(mousebutton)
    if (not InCombatLockdown()) then
	local button = this;
	local classID = button:GetAttribute("classID");
	local spell = PallyPower:GetSpell(classID);
	local gspell = L["SPELL_GTPREF"] .. spell .. L["SPELL_GTSUFF"];
	local unitid = PallyPower:GetUnitSmart(classID, mousebutton, spell);
	if not unitid then
	    spell = "qq";
	    gspell = "qq";
	end
        -- left click (find first nearby player and do 15 minute buff)
	button:SetAttribute("unit1", unitid);
	button:SetAttribute("spell1", gspell);
        -- right click (find first nearby player without buff and do a 5 minute buff)
	button:SetAttribute("unit2", unitid);
	button:SetAttribute("spell2", spell);
        PallyPower.Header:SetAttribute("addchild", button);
    end
end

function PallyPower:ButtonPostClick(mousebutton)
    local button = this;
    local classID = button:GetAttribute("classID");
    if (mousebutton == "LeftButton") then
	LastCast[classID] = 15 * 60;
    end
end

function PallyPower:PopupButtonPreClick(mousebutton)

end

function PallyPower:PopupButtonPostClick(mousebutton)
    local button = this;
    local classID = button:GetAttribute("classID");
    if (mousebutton == "LeftButton") then
	LastCast[classID] = 15 * 60;
    end
end

function PallyPower:DewClick()
    dewdrop:Open(PallyPowerConfigFrame);
end

--
-- Drag Handle
--

-- Lock & Unlock the frame on left click, and toggle config dialog with right click
function PallyPower:ClickHandle(button)
    local function RelockActionBars()
	self.opt.display.frameLocked = true;
    	if (self.opt.display.LockBuffBars) then
		LOCK_ACTIONBAR = "1";
	end
    	_G["PallyPowerAnchor"]:SetChecked(true);
    end

    if (button == "RightButton") then
	PallyPowerConfig_Toggle();
	this:SetChecked(self.opt.display.frameLocked);
    elseif (button == "LeftButton") then
	self.opt.display.frameLocked = not self.opt.display.frameLocked;
	if (self.opt.display.frameLocked) then
	    if (self.opt.display.LockBuffBars) then
		LOCK_ACTIONBAR = "1";
	    end
	else
	    if (self.opt.display.LockBuffBars) then
		LOCK_ACTIONBAR = "0";
	    end
	    self:ScheduleEvent("PallyPowerTemporaryUnlock", RelockActionBars, 30);
	end
	this:SetChecked(self.opt.display.frameLocked);
    end
end

-- Start dragging if not locked
function PallyPower:DragStart()
    if (not self.opt.display.frameLocked) then
	_G["PallyPowerFrame"]:StartMoving();
    end
end

-- End dragging
function PallyPower:DragStop()
    _G["PallyPowerFrame"]:StopMovingOrSizing();
end
